package org.processmining.plugins.ppm.ppmchart.model;

import java.util.HashMap;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartPanel;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class BPMNModel {
	private boolean minimum, actualDist, alsoReverse;
	private HashMap<String, Node> nodes;
	private HashMap<String, Edge> edges;
	private HashMap<Node, HashMap<Node, Double>> distInclDel;
	private HashMap<Node, HashMap<Node, String>> pathInclDel;
	private HashMap<Node, HashMap<Node, Double>> distExclDel;
	private HashMap<Node, HashMap<Node, String>> pathExclDel;
	private Node startEvent;

	//minimum:      calculate minimum distance (true) or maximum distance (from start) (false)
	//actualDist:   calculate geographical distance (true) or number of elements (false)
	//alsoReverse:  mind directions of arrows (true) or not (false), when calculating minimum/maximum distance
	public BPMNModel(XLog log, boolean minimum, boolean actualDist, boolean alsoReverse) {
		this.minimum = minimum;
		this.actualDist = actualDist;
		this.alsoReverse = alsoReverse;
		nodes = new HashMap<String, Node>();
		edges = new HashMap<String, Edge>();
		
		distInclDel = new HashMap<Node, HashMap<Node,Double>>();
		pathInclDel = new HashMap<Node, HashMap<Node,String>>();
		distExclDel = new HashMap<Node, HashMap<Node,Double>>();
		pathExclDel = new HashMap<Node, HashMap<Node,String>>();
		
		startEvent = null;
		
		calculateModel(log);
		calculateDistances();
	}
	private void calculateModel(XLog log) {
		for(XTrace trace : log) {
			for (XEvent event : trace) {
				XAttributeMap map = event.getAttributes();
				String activity = XLogFunctions.getName(event).toUpperCase();
				if (!map.containsKey("id"))
					continue;
				String id = map.get("id").toString();
				if (isNode(activity)) {
					long x = map.containsKey("x") ? XLogFunctions.getLongValue(map.get("x")) : 0;
					long y = map.containsKey("y") ? XLogFunctions.getLongValue(map.get("y")) : 0;
					if (isCreate(activity))
						if (nodes.containsKey(id)) {
							nodes.get(id).setType(getType(activity)).setX(x).setY(y).setDeleted(false);
							if (isStartEvent(activity))
								startEvent = nodes.get(id);
						}
						else {
							Node node = new Node(id, getType(activity), x, y);
							if (isStartEvent(activity))
								startEvent = node;
							nodes.put(id, node);
						}
					else if (isMove(activity))
						nodes.get(id).setX(x).setY(y);
					else if (isDelete(activity))
						nodes.get(id).setDeleted(true);
				}
				else if (isEdge(activity)) {
					String source = map.containsKey("source_node_id") ? XLogFunctions.getStringValue(map.get("source_node_id")) : "";
					String target = map.containsKey("target_node_id") ? XLogFunctions.getStringValue(map.get("target_node_id")) : "";
					if (isCreate(activity)) {
						if (!nodes.containsKey(source))
							nodes.put(source, new Node(source, 0, 0));
						if (!nodes.containsKey(target))
							nodes.put(target, new Node(target, 0, 0));
						edges.put(id, new Edge(id, nodes.get(source), nodes.get(target)));
					}
					else if (isReconnect(activity)) {
						Edge edge = edges.get(id).setId("OLD" + id).setDeleted(true);
						edges.put("OLD" + id, edge);
						if (!nodes.containsKey(source))
							nodes.put(source, new Node(source, 0, 0));
						if (!nodes.containsKey(target))
							nodes.put(target, new Node(target, 0, 0));
						edges.put(id, new Edge(id, nodes.get(source), nodes.get(target)));
					}
					else if (isDelete(activity)) {
						edges.get(id).setDeleted(true);
					}
				}
			}
		}
		if (startEvent == null) {
			System.out.println("model" + XLogFunctions.getID(log));
			for (String n : nodes.keySet()) {
				Node node = nodes.get(n);
				if (node.isDeleted())
					continue;
				boolean isTarget = false;
				for (String e : edges.keySet()) {
					Edge edge = edges.get(e);
					if (edge.isDeleted())
						continue;
					if (edge.getTarget().equals(node)) {
						isTarget = true;
						break;
					}
				}
				if (!isTarget) {
					startEvent = node;
					break;
				}
			}
		}
		if (startEvent == null) {
			System.out.println("StartEvent is null??? (model" + XLogFunctions.getID(log) + ")");
		}
	}
	private void calculateDistances() {
		for (String n : nodes.keySet())
		{
			Node node = nodes.get(n);
			HashMap<Node, Double> mapInclDel = new HashMap<Node, Double>();
			HashMap<Node, String> mapInclDelPath = new HashMap<Node, String>();
			HashMap<Node, Double> mapExclDel = new HashMap<Node, Double>();
			HashMap<Node, String> mapExclDelPath = new HashMap<Node, String>();
			for (String n2 : nodes.keySet())
			{
				Node node2 = nodes.get(n2);
				mapInclDel.put(node2, 0.0);
				mapInclDelPath.put(node2, "");
				mapExclDel.put(node2, 0.0);
				mapExclDelPath.put(node2, "");
			}
			distInclDel.put(node, mapInclDel);
			pathInclDel.put(node, mapInclDelPath);
			distExclDel.put(node, mapExclDel);
			pathExclDel.put(node, mapExclDelPath);
		}
		for (String e : edges.keySet())
		{
			Edge edge = edges.get(e);
			Node s = edge.getSource();
			Node t = edge.getTarget();
			String path = "-" + s.getId() + "-" + t.getId() + "-";
			double dist = actualDist ? getDirectDistance(s, t) : 2;
			double currDist = distExclDel.get(s).get(t);
			if (currDist == 0.0 
				|| (minimum && dist < currDist)
				|| (!minimum && dist > currDist))
			{
				distInclDel.get(s).put(t, dist);
				pathInclDel.get(s).put(t, path);
				if (!edge.isDeleted())
				{
					distExclDel.get(s).put(t, dist);
					pathExclDel.get(s).put(t, path);
				}
			}	
			if (alsoReverse)
			{	
				String revPath = "-" + t.getId() + "-" + s.getId() + "-";
				double revDist = actualDist ? getDirectDistance(t, s) : 2;
				double currRevDist = distExclDel.get(t).get(s);
				if (currRevDist == 0.0
					|| (minimum && revDist < currRevDist)
					|| (!minimum && revDist > currRevDist))
				{
					distInclDel.get(t).put(s, revDist);
					pathInclDel.get(t).put(s, revPath);
					if (!edge.isDeleted())
					{
						distExclDel.get(t).put(s, revDist);
						pathExclDel.get(t).put(s, revPath);
					}
				}
			}
		}
		int changes = aggregate(distInclDel, pathInclDel); 
		while(changes > 0)
			changes = aggregate(distInclDel, pathInclDel);
		
		changes = aggregate(distExclDel, pathExclDel); 
		while(changes > 0)
			changes = aggregate(distExclDel, pathExclDel);
	}
	private int aggregate(HashMap<Node, HashMap<Node, Double>> dists, HashMap<Node, HashMap<Node, String>> paths) {
		int changes = 0;
		for (Node s : dists.keySet())
		{
			for (Node t: dists.get(s).keySet())
			{
				if (dists.get(s).get(t) > 0.0)
				{
					changes += aggregate(dists, paths, s, t);
				}
			}
		}
		return changes;
	}
	private int aggregate(HashMap<Node, HashMap<Node, Double>> dists, HashMap<Node, HashMap<Node, String>> paths, Node s, Node t) {
		int changes = 0;
		// loop over nodes that start at t
		for (Node t2 : dists.get(t).keySet()) 
		{
			if (dists.get(t).get(t2) > 0.0)
			{
				double newDist = dists.get(s).get(t) + dists.get(t).get(t2);
				double currDist = dists.get(s).get(t2);
				if (currDist == 0.0 
					|| (minimum && newDist < currDist)
					|| (!minimum && newDist > currDist))
				{
					if (!overlap(paths.get(s).get(t), paths.get(t).get(t2))) 
					{
						dists.get(s).put(t2, newDist);
						String path1 = paths.get(s).get(t);   // last node in path1 is t
						String path2 = paths.get(t).get(t2);  // first node in path2 is t
						String newPath = path1 + path2.substring(path2.indexOf("-",1) + 1); // strip second t
						paths.get(s).put(t2, newPath);
						changes++;
					}
				}
			}

		}
		return changes;
	}
	// -a-b-...-c- and -c-d-...-e-
	private boolean overlap(String path, String path2) {
		String[] ids = path2.split("-");       // result={"","c","d",...,"e",""}
		for (int i = 2; i<ids.length-1; i++)   //    skip "","c"    and skip ""
		{
			if (path.contains("-" + ids[i] + "-"))
				return true;
		}
		return false;
	}
	private double getDirectDistance(Node a, Node b) {
		if (a==null || b == null)
			return Double.MAX_VALUE;
		return Math.sqrt(Math.pow(a.getX() - b.getX(),2) + Math.pow(a.getY() - b.getY(),2));
	}
	
	public double getDistance(String source, String target) {
		if (nodes.containsKey(source)) {
			if (nodes.containsKey(target)) {
				return getDistance(nodes.get(source), nodes.get(target));
			}
			else if (edges.containsKey(target)) {
				Edge edge = edges.get(target);
				double distSource = getDistance(nodes.get(source), edge.getSource());
				double distTarget = getDistance(nodes.get(source), edge.getTarget());
				return (distSource + distTarget) / 2.0;
			}
		}
		else if (edges.containsKey(source)) {
			if (nodes.containsKey(target)) {
				Edge edge = edges.get(source);
				double distSource = getDistance(edge.getSource(), nodes.get(target));
				double distTarget = getDistance(edge.getTarget(), nodes.get(target));
				return (distSource + distTarget) / 2.0;
			}
			else if (edges.containsKey(target)) {
				Edge edgeSource = edges.get(source);
				Edge edgeTarget = edges.get(target);
				double distSource = getDistance(edgeSource.getSource(), edgeTarget.getSource());
				double distTarget = getDistance(edgeSource.getTarget(), edgeTarget.getTarget());
				return (distSource + distTarget) / 2.0;
			}
		}
		return Double.MAX_VALUE;
	}
	public double getDistance(Node source, Node target) {
		if (source == null || target == null)
			System.out.println("piep");
		if (source.equals(target))
			return 0;

		if (distExclDel.containsKey(source)
				&& distExclDel.get(source).containsKey(target)
				&& distExclDel.get(source).get(target) > 0)
			return distExclDel.get(source).get(target);
		
		if (distInclDel.containsKey(source)
			&& distInclDel.get(source).containsKey(target)
			&& distInclDel.get(source).get(target) > 0)
			return distInclDel.get(source).get(target);
		
		return Double.MAX_VALUE;
	}
	
	public double getDistanceFromStart(String id, String sortStandard) {
		if (nodes.containsKey(id))
			return getDistanceFromStart(nodes.get(id), sortStandard);
		else if (edges.containsKey(id))
			return getDistanceFromStart(edges.get(id), sortStandard);
		else
			return Double.MAX_VALUE;
	}
	public double getDistanceFromStart(Node node, String sortStandard) {
		return (getDistance(startEvent, node));
	}
	public double getDistanceFromStart(Edge e, String sortStandard) {
		if (sortStandard.equals(PPMChartPanel.ST_DIST_START))
			return (getDistanceFromStart(e.getSource(), sortStandard) + getDistanceFromStart(e.getTarget(), sortStandard)) / 2.0;
		else
			return Math.max(getDistanceFromStart(e.getSource(), sortStandard), getDistanceFromStart(e.getTarget(), sortStandard) + 0.5);
	}
	
	public int getNumElements() {
		return nodes.size() + edges.size();
	}
	
	
	protected Node.TYPE getType(String activity) {
		if (activity.endsWith("_ACTIVITY"))
			return Node.TYPE.ACTIVITY;
		else if (activity.endsWith("_START_EVENT"))
			return Node.TYPE.START_EVENT;
		else if (activity.endsWith("_END_EVENT"))
			return Node.TYPE.END_EVENT;
		else if (activity.endsWith("_XOR"))
			return Node.TYPE.XOR;
		else if (activity.endsWith("_AND"))
			return Node.TYPE.AND;
		else
			return Node.TYPE.UNKNOWN;
	}
	protected boolean isNode(String activity) {
		return activity.endsWith("_ACTIVITY") ||
			   activity.endsWith("_START_EVENT") ||
			   activity.endsWith("_END_EVENT") ||
			   activity.endsWith("_XOR") ||
			   activity.endsWith("_AND");
	}
	protected boolean isStartEvent(String activity) {
		return activity.endsWith("_START_EVENT");
	}
	protected boolean isEdge(String activity) {
		return activity.endsWith("_EDGE");
	}
	protected boolean isCreate(String activity) {
		return activity.startsWith("CREATE_");
	}
	protected boolean isMove(String activity) {
		return activity.startsWith("MOVE_");
	}
	protected boolean isReconnect(String activity) {
		return activity.startsWith("RECONNECT_");
	}
	protected boolean isDelete(String activity) {
		return activity.startsWith("DELETE_");
	}
}
