package org.processmining.plugins.ppm.ppmchart.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.ppm.statistics.DoubleHashMap;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class ChunkModel {
	private static final int OPERATION_TYPE = 5;
	private static final int ELEMENT_TYPE = 1;
	private static final int LOOKBACK_WINDOW = 1;
	private static final boolean SIMUL_OPERATIONS_IS_ONE = false;
	
	public static final int LOCALITY = 0;
	public static final int SIMILARITY = 1;
	public static final int TIMING = 2;
	public static final int BLOCKS = 3;
	
	private ArrayList<Long> chunkTimes;
	private int numLocality;
	private int numSimilarity;
	private int numTiming;
	private int numBlocks;
	
	private ArrayList<ChunkModelElement> localityElements;
	private ArrayList<ChunkModelElement> similarityElements;
	private ArrayList<ChunkModelElement> timingElements;
	private ArrayList<ChunkModelElement> blockElements;

	private DoubleHashMap x_values = new DoubleHashMap();
	private DoubleHashMap y_values = new DoubleHashMap();
	
	public ChunkModel(XLog log, int numLoc, int numSim, int numTim, int numBlo) {
		numLocality = numLoc;
		numSimilarity = numSim;
		numTiming = numTim;
		numBlocks = numBlo;
		createModel(log);
	}
	public void set(int type, int val) {
		if (type == LOCALITY)
			setNumLocality(val);
		else if (type == SIMILARITY)
			setNumSimilarity(val);
		else if (type == TIMING)
			setNumTiming(val);
		else if (type == BLOCKS)
			setNumBlocks(val);
	}
	public void setNumLocality(int numLoc) {
		numLocality = numLoc;
	}
	public void setNumSimilarity(int numSim) {
		numSimilarity = numSim;
	}
	public void setNumTiming(int numTim) {
		numTiming = numTim;
	}
	public void setNumBlocks(int numBlo) {
		numBlocks = numBlo;
	}
	public ArrayList<ChunkModelElement> getBlockElements() {
		return blockElements;
	}
	private void createModel(XLog log) {
		ArrayList<ComparableXEvent> events = new ArrayList<ComparableXEvent>();
		for(XTrace trace : log) {
			for (XEvent event : trace) {
				if (isEventOfInterest(event))
					events.add(new ComparableXEvent(event));
			}
		}
		Collections.sort(events);
		
		localityElements = new ArrayList<ChunkModelElement>();
		similarityElements = new ArrayList<ChunkModelElement>();
		timingElements = new ArrayList<ChunkModelElement>();
		blockElements = new ArrayList<ChunkModelElement>();
		HashMap<XEvent, Double> blockDistances = prepareBlockDistances(events);
		
		XEvent prevEvent = null;
		ArrayList<String> similarityBenchmarks = new ArrayList<String>();
		for (ComparableXEvent cEvent : events) {
			XEvent event = cEvent.getEvent();
			String activity = XLogFunctions.getName(event);
			register(event);	
			if (prevEvent == null) {
				prevEvent = event;
				similarityBenchmarks.add(activity);
				continue;
			}
					
			long time = (long)((XLogFunctions.getTime(event).getTime() + XLogFunctions.getTime(prevEvent).getTime()) / 2.0); 
			localityElements.add(new ChunkModelElement(ChunkModel.LOCALITY, localityDistance(prevEvent, event), time, prevEvent, event));
			similarityElements.add(new ChunkModelElement(ChunkModel.SIMILARITY, similarityDistance(activity, similarityBenchmarks), time, prevEvent, event));
			timingElements.add(new ChunkModelElement(ChunkModel.TIMING, timingDistance(prevEvent, event), time, prevEvent, event));
			blockElements.add(new ChunkModelElement(ChunkModel.BLOCKS, blockDistances.get(event), time, prevEvent, event));
			
			similarityBenchmarks.add(activity);
			if (similarityBenchmarks.size() > LOOKBACK_WINDOW)
				similarityBenchmarks.remove(0);
			prevEvent = event;
		}
		Collections.sort(localityElements);
		Collections.sort(similarityElements);
		Collections.sort(timingElements);
		Collections.sort(blockElements);
	}
	private boolean isEventOfInterest(XEvent event) {
		return isEventOfIntereset(XLogFunctions.getName(event));
	}
	private double localityDistance(XEvent event1, XEvent event2) {
		double x1 = getXValue(event1);
		double y1 = getYValue(event1);
		double x2 = getXValue(event2);
		double y2 = getYValue(event2);
		return Math.sqrt(Math.pow((x2 - x1),2) + Math.pow((y2 - y1),2));
	}
	private String getId(XEvent event)
	{
		if (event.getAttributes().containsKey("id"))
			return XLogFunctions.getStringValue(event.getAttributes().get("id"));
		else
			return "";
	}
	private long getX(XEvent event)
	{
		if (event.getAttributes().containsKey("x"))
			return XLogFunctions.getLongValue(event.getAttributes().get("x"));
		else
			return -1;
	}
	private long getY(XEvent event)
	{
		if (event.getAttributes().containsKey("y"))
			return XLogFunctions.getLongValue(event.getAttributes().get("y"));
		else
			return -1;
	}
	private void register(XEvent event)
	{
		XAttributeMap attMap = event.getAttributes();
		if (!attMap.containsKey("id") || !attMap.containsKey("x") || !attMap.containsKey("y"))
			return;
					
		x_values.put(getId(event), getX(event));
		y_values.put(getId(event), getY(event));
	}
	private double getXValue(XEvent event)
	{
		if (x_values.containsKey(getId(event)))
		{
			return x_values.get(getId(event));
		}
		else if (XLogFunctions.getName(event).equals("CREATE_EDGE"))
		{
			double x1 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			double x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			return (x1 + x2) / 2.0;
		}
		else if (XLogFunctions.getName(event).equals("RECONNECT_EDGE"))
		{
			if (!event.getAttributes().containsKey("source_node_id"))
				return x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			else if (!event.getAttributes().containsKey("target_node_id"))
				return x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			else if (!event.getAttributes().containsKey("old_node_id"))
			{
				double x1 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				double x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (x1 + x2) / 2.0;
			}
			else
			{
				double x2, x1 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("old_node_id")));
				if (XLogFunctions.getStringValue(event.getAttributes().get("old_node_type")).equalsIgnoreCase("Source"))
					x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				else
					x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (x1 + x2) / 2.0;
			}
		}
		return -1;
	}
	private double getYValue(XEvent event)
	{
		if (y_values.containsKey(getId(event)))
		{
			return y_values.get(getId(event));
		}
		else if (XLogFunctions.getName(event).equals("CREATE_EDGE"))
		{
			double y1 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			double y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			return (y1 + y2) / 2.0;
		}
		else if (XLogFunctions.getName(event).equals("RECONNECT_EDGE"))
		{
			if (!event.getAttributes().containsKey("source_node_id"))
				return y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			else if (!event.getAttributes().containsKey("target_node_id"))
				return y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			else if (!event.getAttributes().containsKey("old_node_id"))
			{
				double y1 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				double y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (y1 + y2) / 2.0;
			}
			else
			{
				double y2, y1 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("old_node_id")));
				if (XLogFunctions.getStringValue(event.getAttributes().get("old_node_type")).equalsIgnoreCase("Source"))
					y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				else
					y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (y1 + y2) / 2.0;
			}
		}
		return -1;
	}
	
	private double similarityDistance(String activity, ArrayList<String> benchmarks)
	{
		double minDistance = -1;
		for (String benchmark : benchmarks)
		{
			double distance = 0;
			if (!getOperationType(activity).equals(getOperationType(benchmark)))
				distance += OPERATION_TYPE;
			if (!getElementType(activity).equals(getElementType(benchmark)))
				distance += ELEMENT_TYPE;
			if (minDistance == -1 || distance < minDistance)
				minDistance = distance;
		}
		return minDistance;
	}
	private String getOperationType(String activity)
	{
		return activity.substring(0, activity.indexOf("_") - 1);
	}
	private String getElementType(String activity)
	{
		return activity.substring(activity.indexOf("_") + 1);
	}
	
	private double timingDistance(XEvent event1, XEvent event2) {
		return XLogFunctions.getTime(event2).getTime() - XLogFunctions.getTime(event1).getTime();
	}
	
	private HashMap<XEvent, Double> prepareBlockDistances (ArrayList<ComparableXEvent> events) {
		HashMap<XEvent, Double> blockDistances = new HashMap<XEvent, Double>();
		ArrayList<String> ids = new ArrayList<String>();
		HashMap<String, Integer> before = new HashMap<String, Integer>();
		HashMap<String, Integer> after = new HashMap<String, Integer>();
		for (ComparableXEvent event : events) {
			String id = getId(event.getEvent());
			before.put(id, 0);
			if(!after.containsKey(id)) {
				ids.add(id);
				after.put(id, 0);
			}
			after.put(id, after.get(id) + 1);
		}
		
		ComparableXEvent prevEvent = null;
		for (ComparableXEvent event : events) {
			int num = 0;
			for (String id : ids)
				if (before.get(id) != 0 && after.get(id) != 0)
					num += Math.min(before.get(id), after.get(id));
			double dist = num;
			if (prevEvent != null)
				dist += (1 - timingDistance(prevEvent.getEvent(), event.getEvent()) / (1 * 60 * 60 * 1000)); // 1h 
			blockDistances.put(event.getEvent(),  dist);
			
			String id = getId(event.getEvent());
			after.put(id, after.get(id) - 1);
			before.put(id, before.get(id) + 1);
			
			prevEvent = event;
		}
		return blockDistances;
	}
	
	public void update() {
		chunkTimes = new ArrayList<Long>();
		int totNumLoc = localityElements.size();
		int totNumSim = similarityElements.size();
		int totNumTim = timingElements.size();
		int totNumBlo = blockElements.size();
		
		for (int i=0; i<Math.min(numLocality, totNumLoc); i++)
			chunkTimes.add(new Long(localityElements.get(/*totNumLoc - 1 -*/ i).getTime()));
		for (int i=0; i<Math.min(numSimilarity, totNumSim); i++)
			chunkTimes.add(new Long(similarityElements.get(/*totNumSim - 1 -*/ i).getTime()));
		for (int i=0; i<Math.min(numTiming, totNumTim); i++)
			chunkTimes.add(new Long(timingElements.get(totNumTim - 1 - i).getTime()));
		for (int i=0; i<Math.min(numBlocks, totNumBlo); i++)
			chunkTimes.add(new Long(blockElements.get(totNumBlo - 1 - i).getTime()));
	}

	public ArrayList<Long> getChunkTimes() {
		update();
		return chunkTimes;
	}
	
	private boolean isActivity(String activity) {
		return activity.toUpperCase().endsWith("_ACTIVITY");
	}
	private boolean isAnd(String activity) {
		return activity.toUpperCase().endsWith("_AND");
	}
	private boolean isXor(String activity) {
		return activity.toUpperCase().endsWith("_XOR");
	}
	private boolean isGateway(String activity) {
		return isAnd(activity) || isXor(activity);
	}
	private boolean isStart(String activity) {
		return activity.toUpperCase().endsWith("_START_EVENT");
	}
	private boolean isEnd(String activity) {
		return activity.toUpperCase().endsWith("_END_EVENT");
	}
	private boolean isStartEnd(String activity) {
		return isStart(activity) || isEnd(activity);
	}
	private boolean isEdge(String activity) {
		return activity.toUpperCase().endsWith("_EDGE");
	}
	private boolean isEdgeBendpoint(String activity) {
		return activity.toUpperCase().endsWith("_EDGE_BENDPOINT");
	}
	private boolean isReconnect(String activity) {
		return activity.toUpperCase().startsWith("RECONNECT_");
	}
	private boolean isEdgeLabel(String activity) {
		return activity.toUpperCase().endsWith("_EDGE_LABEL");
	}
	private boolean isEventOfIntereset(String activity) {
		return isActivity(activity) || isGateway(activity) || isStartEnd(activity) || isEdge(activity) || isEdgeBendpoint(activity); 
	}
}
