package org.processmining.plugins.ppm.statistics;

import java.util.*;

import org.deckfour.xes.model.*;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class FlowNetwork {
	public HashMap<String, ArrayList<CapacityEdge>> nodes = new HashMap<String,ArrayList<CapacityEdge>>();
    public HashMap<CapacityEdge, Integer> flow = new HashMap<CapacityEdge, Integer>();
    public ArrayList<String> gateways = new ArrayList<String>();
    public String start_event = "";
    public String end_event = "";
	
	public FlowNetwork(XTrace trace) {
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event).toUpperCase();
			if (!event.getAttributes().containsKey("id"))
				continue;
			String id = event.getAttributes().get("id").toString();
			if (activity.equals("CREATE_ACTIVITY"))
				addNode(id);
			else if (activity.equals("CREATE_START_EVENT"))
				addStartEvent(id);
			else if (activity.equals("CREATE_END_EVENT"))
				addEndEvent(id);
			else if (activity.equals("CREATE_XOR") || activity.equals("CREATE_AND"))
				addGateway(id);
			else if (activity.equals("CREATE_EDGE")) {
				String source = event.getAttributes().get("source_node_id").toString();
				String target = event.getAttributes().get("target_node_id").toString();
				addEdge(id, source, target);
			}
			else if (activity.equals("DELETE_ACTIVITY"))
				removeNode(id);
			else if (activity.equals("DELETE_START_EVENT"))
				removeStartEvent(id);
			else if (activity.equals("DELETE_END_EVENT"))
				removeEndEvent(id);
			else if (activity.equals("DELETE_XOR") || activity.equals("DELETE_AND"))
				removeGateway(id);
			else if (activity.equals("DELETE_EDGE")) {
				removeEdge(id);
			}
			else if (activity.equals("RECONNECT_EDGE")) {
				try {
					String source = event.getAttributes().get("source_node_id").toString();
					String target = event.getAttributes().get("target_node_id").toString();
					removeEdge(id, source);
					addEdge(id, source, target);
				}
				catch(Exception ex) { //ignore			
				}
			}
		}
    }

	public void addNode(String node) {
		nodes.put(node, new ArrayList<CapacityEdge>());
	}
	public void removeNode(String node) {
        //adj.remove(vertex);
		nodes.put(node, new ArrayList<CapacityEdge>()); // replace current with empty arraylist
	}

	public void addStartEvent(String start) {
		addNode(start);
		start_event = start;
	}
	public void removeStartEvent(String start) {
		removeNode(start);
        start_event = "";
	}
	public void addEndEvent(String end) {
		addNode(end);
		end_event = end;
	}
	public void removeEndEvent(String end) {
		removeNode(end);
        end_event = "";
	}
	
	public void addGateway(String gw) {
		addNode(gw);
		gateways.add(gw);
	}
	public void removeGateway(String gw) {
        removeNode(gw);
        gateways.remove(gw);
	}
 
    public ArrayList<CapacityEdge> getEdges(String node) {
        return nodes.get(node);
    }
 
    public void addEdge(String id, String source, String target) {
        if (source.equals(target))
            System.out.println("ERROR");
        CapacityEdge edge = new CapacityEdge(id,source,target,1);
        CapacityEdge redge = new CapacityEdge(id,target,source,0);
        edge.setRedge(redge);
        redge.setRedge(edge);
        nodes.get(source).add(edge);
        nodes.get(target).add(redge);
        flow.put(edge, new Integer(0));
        flow.put(redge, new Integer(0));
    }
    public void removeEdge(String id) {
    	for (String key : nodes.keySet())
    		for (CapacityEdge edge : nodes.get(key))
    			if (edge.getId().equals(id))
        			{ removeEdge(edge); return; }
    }
    public void removeEdge(String id, String source) {
    	ArrayList<CapacityEdge> edges = nodes.get(source);
    	for (CapacityEdge edge : edges)
    		if (edge.getId().equals(id))
    			{ removeEdge(edge); return; }
    }
    public void removeEdge(CapacityEdge edge) {
    	nodes.get(edge.getSource()).remove(edge);
    	nodes.get(edge.getTarget()).remove(edge.getRedge());
		flow.remove(edge);
		flow.remove(edge.getRedge());
    }
    private void resetFlow() {
    	for (CapacityEdge edge : flow.keySet())
    		flow.put(edge, new Integer(0));
    }
 
    public HashMap<CapacityEdge, Integer> findPath(String source, String target, HashMap<CapacityEdge, Integer> path) {
        if (source.equals(target))
            return path;
        for (CapacityEdge edge : getEdges(source)) {
            int residual = edge.getCapacity() - flow.get(edge).intValue();
            if (residual > 0 && !path.containsKey(edge)) {
            	path.put(edge, residual);
            	HashMap<CapacityEdge, Integer> result = findPath(edge.getTarget(), target, path); 
                if (!result.isEmpty())
                    return result;
                path.remove(edge);
            }
        }
        return new HashMap<CapacityEdge, Integer>();
    }
 
    public int maxFlow(String source, String target) {
    	resetFlow();;
    	HashMap<CapacityEdge, Integer> path = findPath(source, target, new HashMap<CapacityEdge, Integer>());
        while (!path.isEmpty()) {
            int fl = Integer.MAX_VALUE;
            for (Edge edge : path.keySet()) {
            	int res = path.get(edge).intValue();
            	if (res < fl)
            		fl = res;
            }
            for (CapacityEdge edge : path.keySet()) {
                flow.put(edge, flow.get(edge).intValue() + fl);
                flow.put(edge.getRedge(), flow.get(edge.getRedge()).intValue() - fl);
            }
            path = findPath(source, target, new HashMap<CapacityEdge, Integer>());
        }
        int sum = 0;
        for (CapacityEdge edge : getEdges(source))
        	sum += flow.get(edge).intValue();
        return sum;
    }
    
    public CountHashMap findMatchingGateways(String gw) {
		if (!end_event.equals("") && !start_event.equals(""))
			addEdge("-1",end_event,start_event);
    	CountHashMap res = new CountHashMap();
    	for (String gw2 : gateways) {
    		if (gw.equals(gw2))
    			continue;
    		if (maxFlow(gw,gw2) > 1 || maxFlow(gw2,gw) > 1)
    			res.put(gw2, 0);
    	}
    	removeEdge("-1");
    	return res;
    }
    public ArrayList<String> findMatchingBlock(String gw1, String gw2) {
    	ArrayList<String> result = new ArrayList<String>();
    	result.add(gw1);
    	boolean found = addChildren(result,gw1,gw2);
    	result.add(gw2);
    	if (found)
    		return result;
    	else
    		return new ArrayList<String>();
    }
    private boolean addChildren(ArrayList<String> list, String node, String end) {
    	boolean found = false;
    	for (int i=0; i<nodes.get(node).size(); i++) {
    		Edge edge = nodes.get(node).get(i);
			list.add(edge.getId());
			if (edge.getTarget().equals(end))
				found = true;
			else if (!list.contains(edge.getTarget())) {
    			list.add(edge.getTarget());
    			found = found || addChildren(list,edge.getTarget(),end);
    		}
    	}
    	return found;
    }
}
