package org.processmining.plugins.ppm.statistics.ordermodel;

import java.util.ArrayList;
import java.util.HashMap;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class OrderModel {
	private HashMap<String, OrderNode> nodes;
	private HashMap<String, OrderEdge> edges;
	private HashMap<OrderNode, ArrayList<OrderEdge>> edgesOut;
	private OrderNode startNode;
	private int createOrder;
	
	private String modelOrder;
	
	private int difference;
	
	// difference
	// - build end model
	// - determine path with depth-first XOR and breadth-first AND (according to model and when in doubt picking the most favorable option (= with minimum order))
	// - calculate minimal number of alterations to make this the order of the modeler (i.e., -1-2-3-4-...)
	
	
	public OrderModel(XLog log) {
		buildModel(log);
		determineModelOrder();
		determineDifference();
	}
	private void buildModel(XLog log) {
		nodes = new HashMap<String, OrderNode>();
		edges = new HashMap<String, OrderEdge>();
		edgesOut = new HashMap<OrderNode, ArrayList<OrderEdge>>();
		startNode = null;
		createOrder = 1;
		
		for(XTrace trace : log) {
			for (XEvent event : trace) {
				XAttributeMap map = event.getAttributes();
				String activity = XLogFunctions.getName(event).toUpperCase();
				if (!map.containsKey("id"))
					continue;
				String id = map.get("id").toString();
				if (isNode(activity)) {
					if (isCreate(activity)) {
						if (nodes.containsKey(id))
							updateNode(nodes.get(id), getType(activity), createOrder++);
						else
							createAndAddNode(id, getType(activity), createOrder++);
							
						if (isStartEvent(activity))
							startNode = nodes.get(id);
					}
					else if (isDelete(activity)) {
						removeNode(nodes.get(id));
					}
				}
				else if (isEdge(activity)) {
					String source = map.containsKey("source_node_id") ? XLogFunctions.getStringValue(map.get("source_node_id")) : "";
					String target = map.containsKey("target_node_id") ? XLogFunctions.getStringValue(map.get("target_node_id")) : "";
					if (isCreate(activity))
						createAndAddEdge(id, source, target);
					else if (isReconnect(activity)) {
						OrderEdge edge = edges.get(id);
						source = source.equals("") ? edge.getSource().getId() : source;
						target = target.equals("") ? edge.getTarget().getId() : target;
						removeEdge(edge);
						createAndAddEdge(id, source, target);
					}
					else if (isDelete(activity))
						removeEdge(edges.get(id));
				}
			}
		}
		if (startNode == null) {
			for (String n : nodes.keySet()) {
				OrderNode node = nodes.get(n);
				boolean isTarget = false;
				for (String e : edges.keySet()) {
					OrderEdge edge = edges.get(e);
					if (edge.getTarget().equals(node)) {
						isTarget = true;
						break;
					}
				}
				if (!isTarget) {
					startNode = node;
					break;
				}
			}
		}
	}
	private void createAndAddNode(String id, OrderNode.TYPE type, int createOrder) {
		addNode(createNode(id, type, createOrder));
	}
	private OrderNode createNode(String id, OrderNode.TYPE type, int createOrder) {
		return new OrderNode(id, type, createOrder);
	}
	private void addNode(OrderNode node) {
		edgesOut.put(node, new ArrayList<OrderEdge>());
		nodes.put(node.getId(), node);
	}
	private void removeNode(OrderNode node) {
		if (node.equals(startNode))
			startNode = null;
		//edgesOut.remove(node);
		nodes.remove(node.getId());
	}
	private void updateNode(OrderNode node, OrderNode.TYPE type, int createOrder) {
		node.setType(type).setOrder(createOrder);
	}
	
	
	private void createAndAddEdge(String id, String source, String target) {
		addEdge(createEdge(id, source, target));
	}
	private OrderEdge createEdge(String id, String source, String target) {
		if (!nodes.containsKey(source))
			addNode(new OrderNode(source, OrderNode.TYPE.UNKNOWN, -1));
		if (!nodes.containsKey(target))
			addNode(new OrderNode(target, OrderNode.TYPE.UNKNOWN, -1));
		return new OrderEdge(id, nodes.get(source), nodes.get(target), createOrder++);
	}
	private void addEdge(OrderEdge edge) {
		edges.put(edge.getId(), edge);
		edgesOut.get(edge.getSource()).add(edge);
	}
	private void removeEdge(OrderEdge edge) {
		edgesOut.get(edge.getSource()).remove(edge);
		edges.remove(edge.getId());
	}
	
	private void determineModelOrder() {
		modelOrder = "-";
		ArrayList<OrderNode> xorsToDo = new ArrayList<OrderNode>();
		ArrayList<OrderNode> andsToDo = new ArrayList<OrderNode>();
		process(startNode, xorsToDo, andsToDo);
	}
	private void process(OrderNode node, ArrayList<OrderNode> xorsToDo, ArrayList<OrderNode> andsToDo) {
		if (modelOrder.contains("-" + node.getId() + "-")) {
			xorsToDo.remove(node);
			andsToDo.remove(node);
			return;
		}
		
		modelOrder += node.getId() + "-";
		
		if (node.getType().equals(OrderNode.TYPE.AND) && edgesOut.get(node).size() > 0) {
			for (OrderEdge edge : edgesOut.get(node)) {
				andsToDo.add(edge.getTarget());
			}
			OrderNode minimum = null;
			for (OrderNode n : andsToDo)
				if (minimum == null || minimum.getOrder() > n.getOrder())
					minimum = n;
			andsToDo.remove(minimum);
			process(minimum, xorsToDo, andsToDo);
		}
		else if (edgesOut.get(node).size() > 1) { //XOR or (multiple) edges on an activity
			OrderNode minimum = null;
			for (OrderEdge edge : edgesOut.get(node)) {
				if (minimum == null || minimum.getOrder() > edge.getTarget().getOrder())
					minimum = edge.getTarget();
			}
			for (OrderEdge edge : edgesOut.get(node))
				if (!edge.getTarget().equals(minimum))
					xorsToDo.add(edge.getTarget());
			process(minimum, xorsToDo, andsToDo);
		}
		else if (edgesOut.get(node).size() == 1) {
			process(edgesOut.get(node).get(0).getTarget(), xorsToDo, andsToDo);
		}
		else if (andsToDo.size() > 0) {

			OrderNode minimum = null;
			for (OrderNode n : andsToDo)
				if (minimum == null || minimum.getOrder() > n.getOrder())
					minimum = n;
			andsToDo.remove(minimum);
			process(minimum, xorsToDo, andsToDo);
		}
		else if (xorsToDo.size() > 0) {
			OrderNode minimum = null;
			for (OrderNode n : xorsToDo)
				if (minimum == null || minimum.getOrder() > n.getOrder())
					minimum = n;
			xorsToDo.remove(minimum);
			process(minimum, xorsToDo, andsToDo);
		}
	}
	
	private void determineDifference() {
		String usedOrder = "-";
		String[] ids = modelOrder.split("-");
		for (String id : ids)
			if (!id.equals(""))
				usedOrder += nodes.get(id).getOrder() + "-";
		
		String rightOrder = "-";
		String[] orders = usedOrder.split("-");
		for (int i = 1; i<orders.length - 1; i++) {
			int minimum = i;
			for (int j = i+1; j<orders.length - 1; j++) {
				if (Integer.parseInt(orders[j]) < Integer.parseInt(orders[minimum])) {
					minimum = j;
				}
			}
			String temp = orders[i];
			orders[i] = orders[minimum];
			orders[minimum] = temp;
		}
		for (String order : orders)
			if (!order.equals(""))
				rightOrder += order + "-";
		
		String[] usedOrderArray = usedOrder.substring(1, usedOrder.length() - 2).split("-");
		String[] rightOrderArray = rightOrder.substring(1, rightOrder.length() - 2).split("-");
		difference = LevenshteinDistance(usedOrderArray, rightOrderArray);
		
		System.out.println("usedOrder: " + usedOrder);
		System.out.println("rightOrderArray: " + rightOrder);
		System.out.println("Difference: " + difference);
		System.out.println("---");
	}
	
	// adapted from http://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance
	// see also http://en.wikipedia.org/wiki/Levenshtein_distance
	private int LevenshteinDistance (String[] s0, String[] s1) {                          
	    int len0 = s0.length + 1;                                                     
	    int len1 = s1.length + 1;                                                     
	 
	    // the array of distances                                                       
	    int[] cost = new int[len0];                                                     
	    int[] newcost = new int[len0];                                                  
	 
	    // initial cost of skipping prefix in String s0                                 
	    for (int i = 0; i < len0; i++) cost[i] = i;                                     
	 
	    // dynamicaly computing the array of distances                                  
	 
	    // transformation cost for each letter in s1                                    
	    for (int j = 1; j < len1; j++) {                                                
	        // initial cost of skipping prefix in String s1                             
	        newcost[0] = j;                                                             
	 
	        // transformation cost for each letter in s0                                
	        for (int i = 1; i < len0; i++) {                                             
	            // matching current letters in both strings                             
	            int match = (s0[i - 1].equals(s1[j - 1])) ? 0 : 1;             
	 
	            // computing cost for each transformation                               
	            int cost_replace = cost[i - 1] + match;                                 
	            int cost_insert  = cost[i] + 1;                                         
	            int cost_delete  = newcost[i - 1] + 1;                                  
	 
	            // keep minimum cost                                                    
	            newcost[i] = Math.min(Math.min(cost_insert, cost_delete), cost_replace);
	        }                                                                           
	 
	        // swap cost/newcost arrays                                                 
	        int[] swap = cost; cost = newcost; newcost = swap;                          
	    }                                                                               
	 
	    // the distance is the cost for transforming all letters in both strings        
	    return cost[len0 - 1];                                                          
	}
	public int getDifference() {
		return difference;
	}
	
	protected OrderNode.TYPE getType(String activity) {
		if (activity.endsWith("_ACTIVITY"))
			return OrderNode.TYPE.ACTIVITY;
		else if (activity.endsWith("_START_EVENT"))
			return OrderNode.TYPE.START_EVENT;
		else if (activity.endsWith("_END_EVENT"))
			return OrderNode.TYPE.END_EVENT;
		else if (activity.endsWith("_XOR"))
			return OrderNode.TYPE.XOR;
		else if (activity.endsWith("_AND"))
			return OrderNode.TYPE.AND;
		else
			return OrderNode.TYPE.UNKNOWN;
	}
	protected boolean isNode(String activity) {
		return activity.endsWith("_ACTIVITY") ||
			   activity.endsWith("_START_EVENT") ||
			   activity.endsWith("_END_EVENT") ||
			   activity.endsWith("_XOR") ||
			   activity.endsWith("_AND");
	}
	protected boolean isStartEvent(String activity) {
		return activity.endsWith("_START_EVENT");
	}
	protected boolean isEdge(String activity) {
		return activity.endsWith("_EDGE");
	}
	protected boolean isCreate(String activity) {
		return activity.startsWith("CREATE_");
	}
	protected boolean isMove(String activity) {
		return activity.startsWith("MOVE_");
	}
	protected boolean isReconnect(String activity) {
		return activity.startsWith("RECONNECT_");
	}
	protected boolean isDelete(String activity) {
		return activity.startsWith("DELETE_");
	}
}
